@@ -13,13 +13,12 @@ import (
 	"fmt"
 	"net"
 	"net/http"
-	"time"
 )
 
 // Device object brings all parts together, namely:
-//   * HTTP proxy server
-//   * USB-backed http.Transport
-//   * DNS-SD advertiser
+//   - HTTP proxy server
+//   - USB-backed http.Transport
+//   - DNS-SD advertiser
 //
 // There is one instance of Device object per USB device
 type Device struct {
@@ -45,6 +44,11 @@ func NewDevice(desc UsbDeviceDesc) (*Device, error) {
 	var dnssdName string
 	var dnssdServices DNSSdServices
 	var log *LogMessage
+	var hwid string
+	var quirks Quirks
+	var httpstatus int
+	var canPrint bool
+	var canScan bool
 
 	// Create USB transport
 	dev.UsbTransport, err = NewUsbTransport(desc)
@@ -52,10 +56,19 @@ func NewDevice(desc UsbDeviceDesc) (*Device, error) {
 		goto ERROR
 	}
 
+	// Obtain quirks
+	quirks = dev.UsbTransport.Quirks()
+
 	// Obtain device's logger
-	info = dev.UsbTransport.UsbDeviceInfo()
 	dev.Log = dev.UsbTransport.Log()
 
+	// Obtain device info and derived information.
+	info = dev.UsbTransport.UsbDeviceInfo()
+	hwid = fmt.Sprintf("%4.4x&%4.4x", info.Vendor, info.Product)
+
+	canPrint = info.BasicCaps&UsbIppBasicCapsPrint != 0
+	canScan = info.BasicCaps&UsbIppBasicCapsScan != 0
+
 	// Load persistent state
 	dev.State = LoadDevState(info.Ident(), info.Comment())
 
@@ -70,29 +83,40 @@ func NewDevice(desc UsbDeviceDesc) (*Device, error) {
 		goto ERROR
 	}
 
-	// Create HTTP server
-	dev.UsbTransport.SetDeadline(
-		time.Now().
-			Add(DevInitTimeout).
-			Add(dev.UsbTransport.Quirks().GetInitDelay()))
+	// Configure transport for init
+	dev.UsbTransport.SetTimeout(quirks.GetInitTimeout())
 
+	// Create HTTP server
 	dev.HTTPProxy = NewHTTPProxy(dev.Log, listener, dev.UsbTransport)
 
 	// Obtain DNS-SD info for IPP
 	log = dev.Log.Begin()
 	defer log.Commit()
 
-	ippinfo, err = IppService(log, &dnssdServices,
+	ippinfo, httpstatus, err = IppService(log, &dnssdServices,
 		dev.State.HTTPPort, info, dev.UsbTransport.Quirks(),
 		dev.HTTPClient)
 
 	if err != nil {
 		dev.Log.Error('!', "IPP: %s", err)
+
+		canRetry := httpstatus != 0 || ErrIsEOF(err)
+		if canRetry && canPrint && quirks.GetInitRetryPartial() {
+			dev.Log.Begin().
+				Info(' ', "Printer not ready (HTTP status %d)",
+					httpstatus).
+				Info(' ', "Retrying due to the %q quirk",
+					QuirkNmInitRetryPartial).
+				Commit()
+
+			err = ErrPartialInit
+			goto ERROR
+		}
 	}
 
 	log.Flush()
 
-	if dev.UsbTransport.DeadlineExpired() {
+	if dev.UsbTransport.TimeoutExpired() {
 		err = ErrInitTimedOut
 		goto ERROR
 	}
@@ -112,16 +136,29 @@ func NewDevice(desc UsbDeviceDesc) (*Device, error) {
 	}
 
 	// Obtain DNS-SD info for eSCL
-	err = EsclService(log, &dnssdServices, dev.State.HTTPPort, info,
+	httpstatus, err = EsclService(log, &dnssdServices, dev.State.HTTPPort, info,
 		ippinfo, dev.HTTPClient)
 
 	if err != nil {
 		dev.Log.Error('!', "ESCL: %s", err)
+
+		canRetry := httpstatus != 0 || ErrIsEOF(err)
+		if canRetry && canScan && quirks.GetInitRetryPartial() {
+			dev.Log.Begin().
+				Info(' ', "Scanner not ready (HTTP status %d)",
+					httpstatus).
+				Info(' ', "Retrying due to the %q quirk",
+					QuirkNmInitRetryPartial).
+				Commit()
+
+			err = ErrPartialInit
+			goto ERROR
+		}
 	}
 
 	log.Flush()
 
-	if dev.UsbTransport.DeadlineExpired() {
+	if dev.UsbTransport.TimeoutExpired() {
 		err = ErrInitTimedOut
 		goto ERROR
 	}
@@ -150,6 +187,15 @@ func NewDevice(desc UsbDeviceDesc) (*Device, error) {
 		goto ERROR
 	}
 
+	// Add common TXT records:
+	//   - usb_SER=VCF9192281  ; Device USB serial number
+	//   - usb_HWID=0482&069d  ; Its vendor and device ID
+	for i := range dnssdServices {
+		svc := &dnssdServices[i]
+		svc.Txt.Add("usb_SER", info.SerialNumber)
+		svc.Txt.Add("usb_HWID", hwid)
+	}
+
 	// Advertise Web service. Assume it always exists
 	dnssdServices.Add(DNSSdSvcInfo{Type: "_http._tcp", Port: dev.State.HTTPPort})
 
@@ -170,7 +216,7 @@ func NewDevice(desc UsbDeviceDesc) (*Device, error) {
 	})
 
 	// Enable handling incoming requests
-	dev.UsbTransport.SetDeadline(time.Time{})
+	dev.UsbTransport.SetTimeout(0)
 	dev.HTTPProxy.Enable()
 
 	// Start DNS-SD publisher
@@ -198,7 +244,12 @@ ERROR:
 	}
 
 	if dev.UsbTransport != nil {
-		dev.UsbTransport.Close(true)
+		reset := true
+		switch err {
+		case ErrUnusable, ErrPartialInit:
+			reset = false
+		}
+		dev.UsbTransport.Close(reset)
 	}
 
 	if listener != nil {
