@@ -10,6 +10,8 @@ package main
 
 import (
 	"errors"
+	"io"
+	"net/url"
 )
 
 // Error values for ipp-usb
@@ -22,4 +24,15 @@ var (
 	ErrUnusable     = errors.New("Device doesn't implement print or scan service")
 	ErrNoIppUsb     = errors.New("ipp-usb daemon not running")
 	ErrAccess       = errors.New("Access denied")
+	ErrPartialInit  = errors.New("Some parts of device not ready yet")
 )
+
+// ErrIsEOF tells if error is io.EOF, possibly wrapped by
+// the Go HTTP client.
+func ErrIsEOF(err error) bool {
+	if urlerr, ok := err.(*url.Error); ok {
+		return urlerr.Err == io.EOF
+	}
+
+	return err == io.EOF
+}
