@@ -15,13 +15,15 @@ import (
 	"net"
 	"net/http"
 	"sort"
+	"strconv"
 	"sync"
 )
 
 // statusOfDevice represents a status of the particular device
 type statusOfDevice struct {
-	desc UsbDeviceDesc // Device descriptor
-	init error         // Initialization error, nil if none
+	desc     UsbDeviceDesc // Device descriptor
+	init     error         // Initialization error, nil if none
+	HTTPPort int           // Assigned http port for the device
 }
 
 var (
@@ -74,6 +76,7 @@ func StatusFormat() []byte {
 	i := 0
 	for _, status := range statusTable {
 		devs[i] = status
+		i++
 	}
 
 	sort.Slice(devs, func(i, j int) bool {
@@ -86,20 +89,26 @@ func StatusFormat() []byte {
 		buf.WriteString(" not found\n")
 	} else {
 		buf.WriteString("\n")
-		fmt.Fprintf(buf, " Num  Device              Vndr:Prod  Model\n")
+		fmt.Fprintf(buf, " Num  Device              Vndr:Prod  Port  Model\n")
 		for i, status := range devs {
 			info, _ := status.desc.GetUsbDeviceInfo()
 
-			fmt.Fprintf(buf, " %3d. %s  %4.4x:%.4x  %q\n",
-				i+1, status.desc.UsbAddr,
-				info.Vendor, info.Product, info.MfgAndProduct)
+			s := "-"
+			if status.HTTPPort != 0 {
+				s = strconv.Itoa(status.HTTPPort)
+			}
 
-			s := "OK"
+			fmt.Fprintf(buf, " %3d. %s  %4.4x:%.4x  %-5s %q\n",
+				i+1, status.desc.UsbAddr,
+				info.Vendor, info.Product, s,
+				info.MfgAndProduct)
+
+			s = "OK"
 			if status.init != nil {
 				s = devs[i].init.Error()
 			}
 
-			fmt.Fprintf(buf, "      status: %s", s)
+			fmt.Fprintf(buf, "      status: %s\n", s)
 		}
 	}
 
@@ -108,11 +117,12 @@ func StatusFormat() []byte {
 
 // StatusSet adds device to the status table or updates status
 // of the already known device
-func StatusSet(addr UsbAddr, desc UsbDeviceDesc, init error) {
+func StatusSet(addr UsbAddr, desc UsbDeviceDesc, HTTPPort int, init error) {
 	statusLock.Lock()
 	statusTable[addr] = &statusOfDevice{
-		desc: desc,
-		init: init,
+		desc:     desc,
+		init:     init,
+		HTTPPort: HTTPPort,
 	}
 	statusLock.Unlock()
 }
