@@ -9,6 +9,7 @@
 package main
 
 import (
+	"context"
 	"encoding/binary"
 	"errors"
 	"runtime"
@@ -23,12 +24,14 @@ import (
 //
 // int libusbHotplugCallback (libusb_context *ctx, libusb_device *device,
 //     libusb_hotplug_event event, void *user_data);
+// void libusbTransferCallback (struct libusb_transfer *transfer);
 //
 // typedef struct libusb_device_descriptor libusb_device_descriptor_struct;
 // typedef struct libusb_config_descriptor libusb_config_descriptor_struct;
 // typedef struct libusb_interface libusb_interface_struct;
 // typedef struct libusb_interface_descriptor libusb_interface_descriptor_struct;
 // typedef struct libusb_endpoint_descriptor libusb_endpoint_descriptor_struct;
+// typedef struct libusb_transfer libusb_transfer_struct;
 //
 // // Note, libusb_strerror accepts enum libusb_error argument, which
 // // unfortunately behaves differently depending on target OS and compiler
@@ -57,7 +60,7 @@ type UsbErrCode int
 
 // UsbErrCode constants
 const (
-	UsbIO            UsbErrCode = C.LIBUSB_ERROR_IO
+	UsbEIO           UsbErrCode = C.LIBUSB_ERROR_IO
 	UsbEInval                   = C.LIBUSB_ERROR_INVALID_PARAM
 	UsbEAccess                  = C.LIBUSB_ERROR_ACCESS
 	UsbENoDev                   = C.LIBUSB_ERROR_NO_DEVICE
@@ -89,6 +92,20 @@ var (
 	// Nonzero, if libusbContextPtr initialized
 	libusbContextOk int32
 
+	// libusbTransferDoneMap contains a map of completion channels,
+	// associated with each active libusb_transfer.
+	//
+	// The libusbTransferCallback uses this map to indicate transfer
+	// completion
+	//
+	// This is required, because CGo is very restrictive in whatever
+	// can be saved in pointer passed to the C side.
+	libusbTransferDoneMap = make(map[*C.libusb_transfer_struct]chan struct{})
+
+	// libusbTransferDoneLock protects multithreaded access to
+	// the libusbTransferDoneMap
+	libusbTransferDoneLock sync.Mutex
+
 	// UsbHotPlugChan receives USB hotplug event notifications
 	UsbHotPlugChan = make(chan struct{}, 1)
 )
@@ -131,7 +148,7 @@ func libusbContext(nopnp bool) (*C.libusb_context, error) {
 		)
 	}
 
-	// Start libusb thread (required for hotplug)
+	// Start libusb thread (required for hotplug and asynchronous I/O)
 	go func() {
 		runtime.LockOSThread()
 		for {
@@ -169,6 +186,99 @@ func libusbHotplugCallback(ctx *C.libusb_context, dev *C.libusb_device,
 	return 0
 }
 
+// Called by libusb on libusb_transfer completion
+//
+//export libusbTransferCallback
+func libusbTransferCallback(xfer *C.libusb_transfer_struct) {
+	// Obtain signaling channel
+	libusbTransferDoneLock.Lock()
+	done := libusbTransferDoneMap[xfer]
+	libusbTransferDoneLock.Unlock()
+
+	// Indicate transfer completion by closing the channel
+	close(done)
+}
+
+// libusbTransferStatusDecode decodes libusb_transfer completion status.
+//
+// It returns either non-negative actual transfer length or error.
+//
+// When computing an error, it consults context.Context cancellation
+// and expiration status.
+func libusbTransferStatusDecode(ctx context.Context,
+	xfer *C.libusb_transfer_struct) (int, error) {
+
+	var rc C.int
+	switch xfer.status {
+	// Handle special cases
+	case C.LIBUSB_TRANSFER_COMPLETED:
+		// Successful completion. Return no error regardless
+		// of the context.Context status.
+		return int(xfer.actual_length), nil
+
+	case C.LIBUSB_TRANSFER_CANCELLED:
+		switch {
+		case ctx.Err() != nil:
+			return 0, ctx.Err()
+		default:
+			rc = C.LIBUSB_ERROR_IO
+		}
+
+	case C.LIBUSB_TRANSFER_TIMED_OUT:
+		// There may be a race between context.Context
+		// expiration and libusb timeout. Be consistent
+		// in returned error.
+		return 0, context.DeadlineExceeded
+
+	// Handle other cases
+	case C.LIBUSB_TRANSFER_STALL:
+		rc = C.LIBUSB_ERROR_PIPE
+
+	case C.LIBUSB_TRANSFER_OVERFLOW:
+		rc = C.LIBUSB_ERROR_OVERFLOW
+
+	case C.LIBUSB_TRANSFER_NO_DEVICE:
+		rc = C.LIBUSB_ERROR_NO_DEVICE
+
+	case C.LIBUSB_TRANSFER_ERROR:
+		rc = C.LIBUSB_ERROR_IO
+
+	default:
+		rc = C.LIBUSB_ERROR_OTHER
+	}
+
+	return 0, UsbError{"libusb_submit_transfer", UsbErrCode(rc)}
+}
+
+// libusbTransferAlloc allocates a libusb_transfer.
+//
+// On success, it allocates a completion channel as well and adds
+// it into the libusbTransferDoneMap.
+func libusbTransferAlloc() (*C.libusb_transfer_struct, chan struct{}, error) {
+	xfer := C.libusb_alloc_transfer(0)
+	if xfer == nil {
+		return nil, nil, UsbError{"libusb_alloc_transfer", UsbENomem}
+	}
+
+	doneChan := make(chan struct{})
+
+	libusbTransferDoneLock.Lock()
+	libusbTransferDoneMap[xfer] = doneChan
+	libusbTransferDoneLock.Unlock()
+
+	return xfer, doneChan, nil
+}
+
+// libusbTransferFree removed libusb_transfer from the libusbTransferDoneMap
+// and releases its memory.
+func libusbTransferFree(xfer *C.libusb_transfer_struct) {
+	libusbTransferDoneLock.Lock()
+	delete(libusbTransferDoneMap, xfer)
+	libusbTransferDoneLock.Unlock()
+
+	C.libusb_free_transfer(xfer)
+}
+
 // UsbCheckIppOverUsbDevices returns true if there are some IPP-over-USB devices
 func UsbCheckIppOverUsbDevices() bool {
 	descs, _ := UsbGetIppOverUsbDeviceDescs()
@@ -520,9 +630,7 @@ func (devhandle *UsbDevHandle) UsbDeviceInfo() (UsbDeviceInfo, error) {
 
 // usbIppBasicCaps reads and decodes printer's
 // Class-specific Device Info Descriptor to obtain device
-// capabilities
-//
-// See IPP USB specification, section 4.3 for details
+// capabilities; see IPP USB specification, section 4.3 for details
 //
 // This function never fails. In a case of errors, it fall backs
 // to the reasonable default
@@ -567,8 +675,8 @@ func (devhandle *UsbDevHandle) usbIppBasicCaps() (caps UsbIppBasicCaps) {
 }
 
 // OpenUsbInterface opens an interface
-func (devhandle *UsbDevHandle) OpenUsbInterface(addr UsbIfAddr) (
-	*UsbInterface, error) {
+func (devhandle *UsbDevHandle) OpenUsbInterface(addr UsbIfAddr,
+	quirks Quirks) (*UsbInterface, error) {
 
 	// Claim the interface
 	rc := C.libusb_claim_interface(
@@ -597,6 +705,7 @@ func (devhandle *UsbDevHandle) OpenUsbInterface(addr UsbIfAddr) (
 	return &UsbInterface{
 		devhandle: devhandle,
 		addr:      addr,
+		quirks:    quirks,
 	}, nil
 }
 
@@ -604,6 +713,7 @@ func (devhandle *UsbDevHandle) OpenUsbInterface(addr UsbIfAddr) (
 type UsbInterface struct {
 	devhandle *UsbDevHandle // Device handle
 	addr      UsbIfAddr     // Interface address
+	quirks    Quirks        // Device quirks
 }
 
 // Close the interface
@@ -619,12 +729,12 @@ func (iface *UsbInterface) Close() {
 //
 // This code was inspired by CUPS, and the original comment follows:
 //
-//    This soft reset is specific to the printer device class and is much less
-//    invasive than the general USB reset libusb_reset_device(). Especially it
-//    does never happen that the USB addressing and configuration changes. What
-//    is actually done is that all buffers get flushed and the bulk IN and OUT
-//    pipes get reset to their default states. This clears all stall conditions.
-//    See http://cholla.mmto.org/computers/linux/usb/usbprint11.
+//	This soft reset is specific to the printer device class and is much less
+//	invasive than the general USB reset libusb_reset_device(). Especially it
+//	does never happen that the USB addressing and configuration changes. What
+//	is actually done is that all buffers get flushed and the bulk IN and OUT
+//	pipes get reset to their default states. This clears all stall conditions.
+//	See http://cholla.mmto.org/computers/linux/usb/usbprint11.
 func (iface *UsbInterface) SoftReset() error {
 	rc := C.libusb_control_transfer(
 		(*C.libusb_device_handle)(iface.devhandle),
@@ -651,24 +761,61 @@ func (iface *UsbInterface) SoftReset() error {
 
 // Send data to interface. Returns count of bytes actually transmitted
 // and error, if any
-func (iface *UsbInterface) Send(data []byte,
-	timeout time.Duration) (n int, err error) {
+func (iface *UsbInterface) Send(ctx context.Context,
+	data []byte) (n int, err error) {
 
-	var transferred C.int
+	// Don't even bother to send, if context already expired
+	if ctx.Err() != nil {
+		return 0, ctx.Err()
+	}
 
-	rc := C.libusb_bulk_transfer(
+	// Allocate a libusb_transfer.
+	xfer, doneChan, err := libusbTransferAlloc()
+	if err != nil {
+		return
+	}
+
+	defer libusbTransferFree(xfer)
+
+	// Setup bulk transfer
+	C.libusb_fill_bulk_transfer(
+		xfer,
 		(*C.libusb_device_handle)(iface.devhandle),
 		C.uint8_t(iface.addr.Out|C.LIBUSB_ENDPOINT_OUT),
 		(*C.uchar)(unsafe.Pointer(&data[0])),
 		C.int(len(data)),
-		&transferred,
-		C.uint(timeout/time.Millisecond),
+		C.libusb_transfer_cb_fn(unsafe.Pointer(C.libusbTransferCallback)),
+		nil,
+		0,
 	)
 
+	if iface.quirks.GetZlpSend() {
+		xfer.flags |= C.LIBUSB_TRANSFER_ADD_ZERO_PACKET
+	}
+
+	// Submit transfer
+	rc := C.libusb_submit_transfer(xfer)
 	if rc < 0 {
-		err = UsbError{"libusb_bulk_transfer", UsbErrCode(rc)}
+		return 0, UsbError{"libusb_submit_transfer", UsbErrCode(rc)}
+	}
+
+	// Wait for completion
+	select {
+	case <-ctx.Done():
+		C.libusb_cancel_transfer(xfer)
+	case <-doneChan:
+	}
+
+	<-doneChan
+	n, err = libusbTransferStatusDecode(ctx, xfer)
+
+	// Introduce inter-URB send delay, if configured
+	if delay := iface.quirks.GetUsbSendDelay(); delay != 0 {
+		threshold := int(iface.quirks.GetUsbSendDelayThreshold())
+		if n > threshold {
+			time.Sleep(delay)
+		}
 	}
-	n = int(transferred)
 
 	return
 }
@@ -678,10 +825,13 @@ func (iface *UsbInterface) Send(data []byte,
 //
 // Note, if data size is not 512-byte aligned, and device has more data,
 // that fits the provided buffer, LIBUSB_ERROR_OVERFLOW error may occur
-func (iface *UsbInterface) Recv(data []byte,
-	timeout time.Duration) (n int, err error) {
+func (iface *UsbInterface) Recv(ctx context.Context,
+	data []byte) (n int, err error) {
 
-	var transferred C.int
+	// Don't even bother to recv, if context already expired
+	if ctx.Err() != nil {
+		return 0, ctx.Err()
+	}
 
 	// Some versions of Linux kernel don't allow bulk transfers to
 	// be larger that 16kb per URB, and libusb uses some smart-ass
@@ -694,23 +844,43 @@ func (iface *UsbInterface) Recv(data []byte,
 		data = data[0:MaxBulkRead]
 	}
 
-	rc := C.libusb_bulk_transfer(
+	// Allocate a libusb_transfer.
+	xfer, doneChan, err := libusbTransferAlloc()
+	if err != nil {
+		return
+	}
+
+	defer libusbTransferFree(xfer)
+
+	// Setup bulk transfer
+	C.libusb_fill_bulk_transfer(
+		xfer,
 		(*C.libusb_device_handle)(iface.devhandle),
 		C.uint8_t(iface.addr.In|C.LIBUSB_ENDPOINT_IN),
 		(*C.uchar)(unsafe.Pointer(&data[0])),
 		C.int(len(data)),
-		&transferred,
-		C.uint(timeout/time.Millisecond),
+		C.libusb_transfer_cb_fn(unsafe.Pointer(C.libusbTransferCallback)),
+		nil,
+		0,
 	)
 
-	if rc == C.LIBUSB_ERROR_PIPE {
-		iface.ClearHalt(true)
+	// Submit transfer
+	rc := C.libusb_submit_transfer(xfer)
+	if rc < 0 {
+		return 0, UsbError{"libusb_submit_transfer", UsbErrCode(rc)}
 	}
 
-	if rc < 0 {
-		err = UsbError{"libusb_bulk_transfer", UsbErrCode(rc)}
+	C.libusb_interrupt_event_handler(libusbContextPtr)
+
+	// Wait for completion
+	select {
+	case <-ctx.Done():
+		C.libusb_cancel_transfer(xfer)
+	case <-doneChan:
 	}
-	n = int(transferred)
+
+	<-doneChan
+	n, err = libusbTransferStatusDecode(ctx, xfer)
 
 	return
 }
