@@ -20,7 +20,6 @@ import (
 	"os"
 	"sort"
 	"strconv"
-	"strings"
 	"sync/atomic"
 	"time"
 
@@ -29,17 +28,18 @@ import (
 
 // UsbTransport implements HTTP transport functionality over USB
 type UsbTransport struct {
-	addr         UsbAddr       // Device address
-	info         UsbDeviceInfo // USB device info
-	log          *Logger       // Device's own logger
-	dev          *UsbDevHandle // Underlying USB device
-	connPool     chan *usbConn // Pool of idle connections
-	connList     []*usbConn    // List of all connections
-	connReleased chan struct{} // Signalled when connection released
-	shutdown     chan struct{} // Closed by Shutdown()
-	connstate    *usbConnState // Connections state tracker
-	quirks       QuirksSet     // Device quirks
-	deadline     time.Time     // Deadline for requests
+	addr           UsbAddr       // Device address
+	info           UsbDeviceInfo // USB device info
+	log            *Logger       // Device's own logger
+	dev            *UsbDevHandle // Underlying USB device
+	connPool       chan *usbConn // Pool of idle connections
+	connList       []*usbConn    // List of all connections
+	connReleased   chan struct{} // Signalled when connection released
+	shutdown       chan struct{} // Closed by Shutdown()
+	connstate      *usbConnState // Connections state tracker
+	quirks         Quirks        // Device quirks
+	timeout        time.Duration // Timeout for requests (0 is none)
+	timeoutExpired uint32        // Atomic non-zero, if timeout expired
 }
 
 // NewUsbTransport creates new http.RoundTripper backed by IPP-over-USB
@@ -71,13 +71,14 @@ func NewUsbTransport(desc UsbDeviceDesc) (*UsbTransport, error) {
 	transport.log.SetLevels(Conf.LogDevice)
 
 	// Setup quirks
-	transport.quirks = Conf.Quirks.ByModelName(transport.info.MfgAndProduct)
+	transport.quirks = Conf.Quirks.MatchByModelName(
+		transport.info.MfgAndProduct)
 
 	// Write device info to the log
 	log := transport.log.Begin().
 		Nl(LogDebug).
 		Debug(' ', "===============================").
-		Info('+', "%s: added %s", transport.addr, transport.info.ProductName).
+		Info('+', "%s: opened %s", transport.addr, transport.info.ProductName).
 		Debug(' ', "Device info:").
 		Debug(' ', "  USB Port:      %d", transport.info.PortNum).
 		Debug(' ', "  Ident:         %s", transport.info.Ident()).
@@ -119,7 +120,7 @@ func NewUsbTransport(desc UsbDeviceDesc) (*UsbTransport, error) {
 	}
 
 	// Hard-reset the device, if needed
-	if transport.quirks.GetInitReset() == QuirksResetHard {
+	if transport.quirks.GetInitReset() == QuirkResetHard {
 		transport.log.Debug(' ', "Doing USB HARD RESET")
 		dev.Reset()
 	}
@@ -172,80 +173,22 @@ ERROR:
 
 // Dump quirks to the UsbTransport's log
 func (transport *UsbTransport) dumpQuirks(log *LogMessage) {
-	type dmp struct {
-		name, val string
-		origin    *Quirks
-	}
-
-	dump := []dmp{}
-
-	var v interface{}
-	var o *Quirks
-
-	// Collect parameters
-	v, o = transport.quirks.GetBlacklistOrigin()
-	dump = append(dump, dmp{"blacklist", fmt.Sprintf("%v", v), o})
-
-	v, o = transport.quirks.GetBuggyIppRspOrigin()
-	dump = append(dump, dmp{"buggy-ipp-responses", fmt.Sprintf("%s", v), o})
-
-	v, o = transport.quirks.GetDisableFaxOrigin()
-	dump = append(dump, dmp{"disable-fax", fmt.Sprintf("%v", v), o})
-
-	v, o = transport.quirks.GetIgnoreIppStatusOrigin()
-	dump = append(dump, dmp{"ignore-ipp-status", fmt.Sprintf("%v", v), o})
-
-	v, o = transport.quirks.GetInitDelayOrigin()
-	dump = append(dump, dmp{"init-delay", fmt.Sprintf("%v", v), o})
-
-	v, o = transport.quirks.GetInitResetOrigin()
-	dump = append(dump, dmp{"init-reset", fmt.Sprintf("%s", v), o})
-
-	v, o = transport.quirks.GetRequestDelayOrigin()
-	dump = append(dump, dmp{"request-delay", fmt.Sprintf("%v", v), o})
-
-	v, o = transport.quirks.GetUsbMaxInterfacesOrigin()
-	dump = append(dump, dmp{"usb-max-interfaces", fmt.Sprintf("%v", v), o})
-
-	// Add HTTP headers
-	for _, q := range transport.quirks {
-		for name, value := range q.HTTPHeaders {
-			d := dmp{
-				name:   "http-" + strings.ToLower(name),
-				val:    fmt.Sprintf("%q", value),
-				origin: q,
-			}
-			dump = append(dump, d)
-		}
-	}
-
-	// Skip no-origin (i.e., actually unset) quirks
-	cnt := 0
-	for _, d := range dump {
-		if d.origin != nil {
-			dump[cnt] = d
-			cnt++
-		}
-	}
-
-	dump = dump[:cnt]
-
-	// Sort by name
-	sort.Slice(dump, func(i, j int) bool {
-		return dump[i].name < dump[j].name
-	})
-
-	// And write to the log
 	log.Debug(' ', "Device quirks:")
 
-	origin := (*Quirks)(nil)
-	for _, d := range dump {
-		if origin != d.origin {
-			origin = d.origin
-			log.Debug(' ', "  from [%s] (%s):",
-				origin.Model, origin.Origin)
+	prevMatch := ""
+	for _, q := range transport.quirks.All() {
+		val := q.RawValue
+		if _, isStr := q.Parsed.(string); isStr {
+			val = strconv.Quote(val)
 		}
-		log.Debug(' ', "    %s = %s", d.name, d.val)
+
+		if q.Match != prevMatch {
+			prevMatch = q.Match
+			log.Debug(' ', "  [%s]", q.Match)
+		}
+
+		log.Debug(' ', "    ; (%s)", q.Origin)
+		log.Debug(' ', "    %s = %s", q.Name, val)
 	}
 }
 
@@ -300,26 +243,21 @@ func (transport *UsbTransport) connInUse() int {
 	return cap(transport.connPool) - len(transport.connPool)
 }
 
-// SetDeadline sets the deadline for all requests, submitted
-// via RoundTrip and RoundTripWithSession methods
-//
-// A deadline is an absolute time after which request processing
-// will fail instead of blocking
+// SetTimeout sets the timeout for all subsequent requests.
 //
 // This is useful only at initialization time and if some requests
 // were failed due to timeout, device reset is required, because
-// at this case synchronization with device will probably be lost
+// at this case synchronization with device will probably be lost.
 //
 // A zero value for t means no timeout
-func (transport *UsbTransport) SetDeadline(t time.Time) {
-	transport.deadline = t
+func (transport *UsbTransport) SetTimeout(t time.Duration) {
+	transport.timeout = t
 }
 
-// DeadlineExpired reports if deadline previously set by SetDeadline()
-// is already expired
-func (transport *UsbTransport) DeadlineExpired() bool {
-	deadline := transport.deadline
-	return !deadline.IsZero() && time.Until(deadline) <= 0
+// TimeoutExpired returns true if one or more of the preceding HTTP request
+// has failed due to timeout.
+func (transport *UsbTransport) TimeoutExpired() bool {
+	return atomic.LoadUint32(&transport.timeoutExpired) != 0
 }
 
 // closeShutdownChan closes the transport.shutdown, which effectively
@@ -387,7 +325,7 @@ func (transport *UsbTransport) Close(reset bool) {
 	}
 
 	transport.dev.Close()
-	transport.log.Info('-', "%s: removed %s",
+	transport.log.Info('-', "%s: closed %s",
 		transport.addr, transport.info.ProductName)
 }
 
@@ -403,7 +341,7 @@ func (transport *UsbTransport) UsbDeviceInfo() UsbDeviceInfo {
 }
 
 // Quirks returns device's quirks
-func (transport *UsbTransport) Quirks() QuirksSet {
+func (transport *UsbTransport) Quirks() Quirks {
 	return transport.quirks
 }
 
@@ -435,13 +373,11 @@ func (transport *UsbTransport) RoundTripWithSession(session int,
 	outreq.Header.Del("Expect")
 
 	// Apply quirks
-	for _, quirks := range transport.quirks {
-		for name, value := range quirks.HTTPHeaders {
-			if value != "" {
-				outreq.Header.Set(name, value)
-			} else {
-				outreq.Header.Del(name)
-			}
+	for name, value := range transport.quirks.HTTPHeaders {
+		if value != "" {
+			outreq.Header.Set(name, value)
+		} else {
+			outreq.Header.Del(name)
 		}
 	}
 
@@ -520,31 +456,66 @@ func (transport *UsbTransport) RoundTripWithSession(session int,
 		time.Sleep(delay)
 	}
 
+	// Set read/write Context. This effectively sets request timeout.
+	//
+	// This is important that context is is set after inter-request
+	// or initial delay is already done, so we don't need to bother
+	// with adjusting the timeout.
+	//
+	// The context cancel function is called from many places and
+	// not always used, so for simplicity I'd better initialize it
+	// to the dummy function rather that to compare it with nil
+	// every time it is called.
+	rwctx := context.Background()
+	cleanupCtx := context.CancelFunc(func() {})
+
+	if transport.timeout != 0 {
+		rwctx, cleanupCtx = context.WithTimeout(rwctx,
+			transport.timeout)
+	}
+
+	conn.setRWCtx(rwctx)
+
 	// Send request and receive a response
 	err = outreq.Write(conn)
 	if err != nil {
 		transport.log.HTTPError('!', session, "%s", err)
 		conn.put()
+		cleanupCtx()
 		return nil, err
 	}
 
 	resp, err := http.ReadResponse(conn.reader, outreq)
 	if err != nil {
+		// If the latest conn.Read has returned io.EOF, the only
+		// reason it could happen is that the zlp-recv-hack
+		// quirk has triggered.
+		//
+		// The stdlib HTTP stack will wrap the io.EOF error into
+		// its own error message. Here we force error condition
+		// back to io.EOF so it cleanly can be detected and handled
+		// by the initialization retry logic at the upper level
+		if conn.EOFSeen() {
+			err = io.EOF
+		}
+
 		transport.log.HTTPError('!', session, "%s", err)
 		conn.put()
+		cleanupCtx()
 		return nil, err
 	}
 
 	// Wrap response body
 	resp.Body = &usbResponseBodyWrapper{
-		log:     transport.log,
-		session: session,
-		body:    resp.Body,
-		conn:    conn,
+		log:        transport.log,
+		session:    session,
+		body:       resp.Body,
+		conn:       conn,
+		cleanupCtx: cleanupCtx,
 	}
 
 	// Optionally sanitize IPP response
-	if transport.quirks.GetBuggyIppRsp() == QuirksBuggyIppRspSanitize &&
+	if transport.quirks.GetBuggyIppRsp() == QuirkBuggyIppRspSanitize &&
 		resp.Header.Get("Content-Type") == "application/ipp" {
 		transport.sanitizeIppResponse(session, resp)
 	}
@@ -650,13 +621,14 @@ func (wrap *usbRequestBodyWrapper) Close() error {
 // usbResponseBodyWrapper wraps http.Response.Body and guarantees
 // that connection will be always drained before closed
 type usbResponseBodyWrapper struct {
-	log     *Logger       // Device's logger
-	session int           // HTTP session, for logging
-	preBody *bytes.Buffer // Data inserted before body, if not nil
-	body    io.ReadCloser // Response.body
-	conn    *usbConn      // Underlying USB connection
-	count   int           // Total count of received bytes
-	drained bool          // EOF or error has been seen
+	log        *Logger            // Device's logger
+	session    int                // HTTP session, for logging
+	preBody    *bytes.Buffer      // Data inserted before body, if not nil
+	body       io.ReadCloser      // Response.body
+	conn       *usbConn           // Underlying USB connection
+	count      int                // Total count of received bytes
+	drained    bool               // EOF or error has been seen
+	cleanupCtx context.CancelFunc // Cancel function for I/O Context
 }
 
 // Read from usbResponseBodyWrapper
@@ -680,8 +652,7 @@ func (wrap *usbResponseBodyWrapper) Read(buf []byte) (int, error) {
 func (wrap *usbResponseBodyWrapper) Close() error {
 	// If EOF or error seen, we can close synchronously
 	if wrap.drained {
-		wrap.body.Close()
-		wrap.conn.put()
+		wrap.cleanup()
 		return nil
 	}
 
@@ -696,28 +667,43 @@ func (wrap *usbResponseBodyWrapper) Close() error {
 		}()
 
 		io.Copy(ioutil.Discard, wrap.body)
-		wrap.body.Close()
-		wrap.conn.put()
+		wrap.cleanup()
 	}()
 
 	return nil
 }
 
+// cleanup performs the final cleanup of the usbResponseBodyWrapper
+// after use.
+func (wrap *usbResponseBodyWrapper) cleanup() {
+	wrap.body.Close()
+	wrap.conn.put()
+
+	// Cleanup I/O context.Context, if any
+	if wrap.cleanupCtx != nil {
+		wrap.cleanupCtx()
+	}
+
+	wrap.log.HTTPDebug('<', wrap.session, "done with response body")
+}
+
 // usbConn implements an USB connection
 type usbConn struct {
-	transport     *UsbTransport // Transport that owns the connection
-	index         int           // Connection index (for logging)
-	iface         *UsbInterface // Underlying interface
-	reader        *bufio.Reader // For http.ReadResponse
-	delayUntil    time.Time     // Delay till this time before next request
-	delayInterval time.Duration // Pause between requests
-	cntRecv       int           // Total bytes received
-	cntSent       int           // Total bytes sent
+	transport     *UsbTransport   // Transport that owns the connection
+	index         int             // Connection index (for logging)
+	iface         *UsbInterface   // Underlying interface
+	reader        *bufio.Reader   // For http.ReadResponse
+	rwctx         context.Context // For usbConn.Read and usbConn.Write
+	delayUntil    time.Time       // Delay till this time before next request
+	delayInterval time.Duration   // Pause between requests
+	cntRecv       int             // Total bytes received
+	cntSent       int             // Total bytes sent
+	eofSeen       bool            // Last usbConn.Read has returned io.EOF
 }
 
 // Open usbConn
 func (transport *UsbTransport) openUsbConn(
-	index int, ifaddr UsbIfAddr, quirks QuirksSet) (*usbConn, error) {
+	index int, ifaddr UsbIfAddr, quirks Quirks) (*usbConn, error) {
 
 	dev := transport.dev
 
@@ -735,13 +721,13 @@ func (transport *UsbTransport) openUsbConn(
 
 	// Obtain interface
 	var err error
-	conn.iface, err = dev.OpenUsbInterface(ifaddr)
+	conn.iface, err = dev.OpenUsbInterface(ifaddr, quirks)
 	if err != nil {
 		goto ERROR
 	}
 
 	// Soft-reset interface, if needed
-	if quirks.GetInitReset() == QuirksResetSoft {
+	if quirks.GetInitReset() == QuirkResetSoft {
 		transport.log.Debug(' ', "USB[%d]: doing SOFT_RESET", index)
 		err = conn.iface.SoftReset()
 		if err != nil {
@@ -762,15 +748,9 @@ ERROR:
 	return nil, err
 }
 
-// Compute Recv/Send timeout
-func (conn *usbConn) timeout() (tm time.Duration, expored bool) {
-	deadline := conn.transport.deadline
-	if deadline.IsZero() {
-		return
-	}
-
-	tm = time.Until(deadline)
-	return tm, tm <= 0
+// setRWCtx sets context.Context for subsequent Read and Write operations
+func (conn *usbConn) setRWCtx(ctx context.Context) {
+	conn.rwctx = ctx
 }
 
 // Read from USB
@@ -778,6 +758,9 @@ func (conn *usbConn) Read(b []byte) (int, error) {
 	conn.transport.connstate.beginRead(conn)
 	defer conn.transport.connstate.doneRead(conn)
 
+	// Drop conn.eofSeenn flag
+	conn.eofSeen = false
+
 	// Note, to avoid LIBUSB_TRANSFER_OVERFLOW errors
 	// from libusb, input buffer size must always
 	// be aligned by 1024 bytes for USB 3.0, 512 bytes
@@ -794,14 +777,14 @@ func (conn *usbConn) Read(b []byte) (int, error) {
 		b = b[0:n]
 	}
 
-	backoff := time.Millisecond * 100
-	for {
-		tm, expired := conn.timeout()
-		if expired {
-			return 0, ErrInitTimedOut
-		}
+	// zlp-recv-hack handling
+	zlpRecvHack := conn.transport.quirks.GetZlpRecvHack()
+	zlpRecv := false
 
-		n, err := conn.iface.Recv(b, tm)
+	// Setup deadline
+	backoff := time.Millisecond * 10
+	for {
+		n, err := conn.iface.Recv(conn.rwctx, b)
 		conn.cntRecv += n
 
 		conn.transport.log.Add(LogTraceHTTP, '<',
@@ -813,16 +796,31 @@ func (conn *usbConn) Read(b []byte) (int, error) {
 		if err != nil {
 			conn.transport.log.Error('!',
 				"USB[%d]: recv: %s", conn.index, err)
+
+			if err == context.DeadlineExceeded {
+				// If we've got read timeout preceded
+				// by the zero-length packet, interpret
+				// is as body EOF condition
+				if zlpRecvHack && zlpRecv {
+					conn.eofSeen = true
+					return 0, io.EOF
+				}
+
+				atomic.StoreUint32(
+					&conn.transport.timeoutExpired, 1)
+			}
 		}
 
 		if n != 0 || err != nil {
 			return n, err
 		}
-		conn.transport.log.Error('!',
+
+		zlpRecv = true
+		conn.transport.log.Debug(' ',
 			"USB[%d]: zero-size read", conn.index)
 
 		time.Sleep(backoff)
-		backoff *= 2
+		backoff += backoff / 4 // The same as backoff *= 1.25
 		if backoff > time.Millisecond*1000 {
 			backoff = time.Millisecond * 1000
 		}
@@ -834,12 +832,7 @@ func (conn *usbConn) Write(b []byte) (int, error) {
 	conn.transport.connstate.beginWrite(conn)
 	defer conn.transport.connstate.doneWrite(conn)
 
-	tm, expired := conn.timeout()
-	if expired {
-		return 0, ErrInitTimedOut
-	}
-
-	n, err := conn.iface.Send(b, tm)
+	n, err := conn.iface.Send(conn.rwctx, b)
 	conn.cntSent += n
 
 	conn.transport.log.Add(LogTraceHTTP, '>',
@@ -851,11 +844,21 @@ func (conn *usbConn) Write(b []byte) (int, error) {
 	if err != nil {
 		conn.transport.log.Error('!',
 			"USB[%d]: send: %s", conn.index, err)
+
+		if err == context.DeadlineExceeded {
+			atomic.StoreUint32(
+				&conn.transport.timeoutExpired, 1)
+		}
 	}
 
 	return n, err
 }
 
+// EOFSeen reports of the latest usbConn.Read has returned io.EOF
+func (conn *usbConn) EOFSeen() bool {
+	return conn.eofSeen
+}
+
 // Allocate a connection
 func (transport *UsbTransport) usbConnGet(ctx context.Context) (*usbConn, error) {
 	select {
